/******************************************************************************
* Make a catalogue of objects for imaging case from a stellar population
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_rng.h>

#define  L_WORD   1000  /* Longest string that can be handled */


char      *progname;


/******************************************************************************
* Print the usage message
******************************************************************************/
 
void usage(void) {

  fprintf(stderr, "\nUsage: %s [options]\n", progname);
  fprintf(stderr, "\nMake a catalogue of objects for the imaging case\n\
from a given stellar population (produced by\n\
rspmkstelpop_basti). Any parameters not specified on\n\
the command line will be queried for.\n");
  fprintf(stderr, "\nOptions include:\n \
\t-help\t\tprint this message\n \
\t-spf file\tstellar population file\n\
\t-dm DM\t\tdistance modulus\n\
\t-sb mu\t\tV band surface brightness\n\
\t-fov om\t\tfield of view size in arcsec\n\
\t-band1 b\tfirst observing band\n\
\t-band2 b\tsecond observing band\n\
\t-ps1 s\t\tpixel size in band 1 in mas\n\
\t-ps2 s\t\tpixel size in band 2 in mas\n\
\t-seed r\t\trandom number seed\n\
\n");
  exit(3);
}


/******************************************************************************
* Main program
******************************************************************************/

int main(int argc, char *argv[]) {

  int        i, j, npts, *star, nstar, brcol;
  long int   seed;
  float      dm, mu, fov, ps1, ps2, *mv, *col11, *col12, *col21, *col22, 
             *m1, *m2;
  double     reqflux, tflux, *w, maxw, *x1, *y1, *x2, *y2, c1, c2;
  char       band1, band2, spf[L_WORD], buffer[L_WORD], *buf;
  FILE       *data_file;
  gsl_rng    *rng;
  float      *readafcol(char *, char *, int *);
  double     *readadcol(char *, char *, int *);
  char       *wdinstr(char *, char *m, int *), *mv2col(char *, int);
  FILE       *faskropen(char *, char *, int);
  void       errormsg(char *, ...), get_input(char *, char *, ...);


  progname = strrchr(argv[0], '/');
  progname = (progname == NULL) ? argv[0] : progname + 1;
  
  /* Look for -help switch */
  for (i = 1; i < argc; i++)
    if (!strcmp(argv[i], "-help"))
      usage();

  /* Some setup */
  strcpy(spf, " ");
  dm = -1.0;
  mu = -1.0;
  fov = -1.0;
  band1 = ' ';
  band2 = ' ';
  ps1 = -1.0;
  ps2 = -1.0;
  seed = 0;
  col11 = NULL;
  col12 = NULL;
  col21 = NULL;
  col22 = NULL;

  /* Look for other switches */
  for (i = 1; i < argc-1; i++)
    if (!strcmp(argv[i], "-spf")) {
      strcpy(spf, argv[++i]);
    } else if (!strcmp(argv[i], "-dm")) {
      if (sscanf(argv[++i], "%f", &dm) != 1 || dm < 0.0)
	errormsg("Bad distance modulus code: %s", argv[i]);
    } else if (!strcmp(argv[i], "-sb")) {
      if (sscanf(argv[++i], "%f", &mu) != 1 || mu < 0.0)
	errormsg("Bad surface brightness: %s", argv[i]);
    } else if (!strcmp(argv[i], "-fov")) {
      if (sscanf(argv[++i], "%f", &fov) != 1 || fov < 0.0)
	errormsg("Bad field of view size: %s", argv[i]);
    } else if (!strcmp(argv[i], "-band1")) {
      if (sscanf(argv[++i], "%c", &band1) != 1 || 
	  (band1 != 'U' && band1 != 'B' && band1 != 'V' && band1 != 'R' && 
	   band1 != 'I' && band1 != 'J' && band1 != 'H' && band1 != 'K' && 
	   band1 != 'L'))
	errormsg("Bad first observing band: %s", argv[i]);
    } else if (!strcmp(argv[i], "-band2")) {
      if (sscanf(argv[++i], "%c", &band2) != 1 || 
	  (band2 != 'U' && band2 != 'B' && band2 != 'V' && band2 != 'R' && 
	   band2 != 'I' && band2 != 'J' && band2 != 'H' && band2 != 'K' && 
	   band2 != 'L'))
	errormsg("Bad second observing band: %s", argv[i]);
    } else if (!strcmp(argv[i], "-ps1")) {
      if (sscanf(argv[++i], "%f", &ps1) != 1 || ps1 < 0.0)
	errormsg("Bad pixel size in band 1: %s", argv[i]);
    } else if (!strcmp(argv[i], "-ps2")) {
      if (sscanf(argv[++i], "%f", &ps2) != 1 || ps2 < 0.0)
	errormsg("Bad pixel size in band 2: %s", argv[i]);
    } else if (!strcmp(argv[i], "-seed")) {
      if (sscanf(argv[++i], "%ld", &seed) != 1)
	errormsg("Bad random number seed: %s", argv[i]);
    } else
      errormsg("Unknown command line option: %s", argv[i]);


  /* Get all missing input */

  /* Stellar population file */
  if (!strcmp(spf, " "))
    i = 1;
  else
    i = 3;
  if ((data_file = faskropen("\nStellar population file?", spf, i)) == NULL)
    errormsg("Couldn't open %s!", spf);

  /* Distance modulus */
  if (dm < 0.0) {
    dm = 31.2;
    do
      get_input("\nDistance modulus?", "%.1f", &dm);
    while (dm < 0.0);
  }

  /* Surface brightness */
  if (mu < 0.0) {
    mu = 22.0;
    do
      get_input("\nSurface brightness (in mag/arcsec^2)?", "%.1f", &mu);
    while (mu < 0.0);
  }

  /* Field of view size */
  if (fov < 0.0) {
    fov = 1.0;
    do
      get_input("\nField of view size (in arcsec)?", "%.1f", &fov);
    while (fov < 0.0);
  }

  /* First observing band */
  if (band1 == ' ') {
    band1 = 'K';
    do
      get_input("\nFirst observing band?", "%c", &band1);
    while (band1 != 'U' && band1 != 'B' && band1 != 'V' && band1 != 'R' && 
	   band1 != 'I' && band1 != 'J' && band1 != 'H' && band1 != 'K' && 
	   band1 != 'L');
  }

  /* Second observing band */
  if (band2 == ' ') {
    band2 = 'I';
    do
      get_input("\nSecond observing band?", "%c", &band2);
    while (band2 != 'U' && band2 != 'B' && band2 != 'V' && band2 != 'R' && 
	   band2 != 'I' && band2 != 'J' && band2 != 'H' && band2 != 'K' && 
	   band2 != 'L');
  }

  /* Pixel size */
  if (ps1 < 0.0) {
    ps1 = 5.0;
    do
      get_input("\nPixel size in band 1 (in mas)?", "%.1f", &ps1);
    while (ps1 < 0.0);
  }

  /* Pixel size */
  if (ps2 < 0.0) {
    ps2 = 5.0;
    do
      get_input("\nPixel size in band 2 (in mas)?", "%.1f", &ps2);
    while (ps2 < 0.0);
  }

  /* Random number seed */
  if (seed == 0)
    seed = time((time_t *)0);

  /* Initialise the random number generator */
  rng = gsl_rng_alloc(gsl_rng_taus2);
  gsl_rng_set(rng, seed);
  fprintf(stderr, "\nRandom number seed = %lu\n", seed);

  /* Read in mags and weights */
  mv = readafcol("Mv", spf, &npts);
  w = readadcol("weight", spf, &npts);
  if (band1 == 'U') {
    col11 = readafcol("(U-B)", spf, &npts);
    col12 = readafcol("(B-V)", spf, &npts);
  } else if (band1 == 'B') {
    col11 = readafcol("(B-V)", spf, &npts);
  } else if (band1 == 'V') {
  } else if (band1 == 'H') {
    col11 = readafcol("(H-K)", spf, &npts);
    col12 = readafcol("(V-K)", spf, &npts);
  } else {
    sprintf(buffer, "(V-%c)", band1);
    col11 = readafcol(buffer, spf, &npts);
  }
  if (band2 == 'U') {
    col21 = readafcol("(U-B)", spf, &npts);
    col22 = readafcol("(B-V)", spf, &npts);
  } else if (band2 == 'B') {
    col21 = readafcol("(B-V)", spf, &npts);
  } else if (band2 == 'V') {
  } else if (band2 == 'H') {
    col21 = readafcol("(H-K)", spf, &npts);
    col22 = readafcol("(V-K)", spf, &npts);
  } else {
    sprintf(buffer, "(V-%c)", band2);
    col21 = readafcol(buffer, spf, &npts);
  }

  /* Requested flux */
  reqflux = pow(10.0, -0.4 * mu) * fov*fov;

  /* Total flux from the stellar population */
  tflux = 0.0;
  for (i = 0; i < npts; i++)
    tflux += pow(10.0, -0.4 * (mv[i] + dm)) * w[i];

  /* Required number of stars */
  nstar = (int)rint(reqflux / tflux);
  fprintf(stderr, "Number of stars = %d\n", nstar);

  /* Make some memory */
  if ((star = (int *)calloc(nstar, sizeof(int))) == NULL)
    errormsg("Couldn't allocate memory!");
  if ((m1 = (float *)calloc(nstar, sizeof(float))) == NULL)
    errormsg("Couldn't allocate memory!");
  if ((m2 = (float *)calloc(nstar, sizeof(float))) == NULL)
    errormsg("Couldn't allocate memory!");
  if ((x1 = (double *)calloc(nstar, sizeof(double))) == NULL)
    errormsg("Couldn't allocate memory!");
  if ((y1 = (double *)calloc(nstar, sizeof(double))) == NULL)
    errormsg("Couldn't allocate memory!");
  if ((x2 = (double *)calloc(nstar, sizeof(double))) == NULL)
    errormsg("Couldn't allocate memory!");
  if ((y2 = (double *)calloc(nstar, sizeof(double))) == NULL)
    errormsg("Couldn't allocate memory!");

  /* Monte Carlo the stars */
  maxw = 1.001 * gsl_stats_max(w, 1, npts);
  for (i = 0; i < nstar; i++) {
    do
      star[i] = gsl_rng_uniform_int(rng, npts);
    while (maxw * gsl_rng_uniform(rng) > w[star[i]]);
  }

  /* Sort the stars; don't add any star properties before this */
  gsl_sort_int(star, 1, nstar);

  /* Actual surface brightness produced by the stars */
  tflux = 0.0;
  for (i = 0; i < nstar; i++)
    tflux += pow(10.0, -0.4 * (mv[star[i]] + dm));
  tflux = -2.5 * log10(tflux / (fov*fov));
  fprintf(stderr, "Actual surface brightness = %.3f mag/arcsec^2\n", tflux);

  /* Apparent magnitudes */
  for (i = 0; i < nstar; i++) {
    if (band1 == 'U') {
      m1[i] = col11[star[i]] + col12[star[i]] + mv[star[i]] + dm;
    } else if (band1 == 'B') {
      m1[i] = col11[star[i]] + mv[star[i]] + dm;
    } else if (band1 == 'V') {
      m1[i] = mv[star[i]] + dm;
    } else if (band1 == 'H') {
      m1[i] = col11[star[i]] - col12[star[i]] + mv[star[i]] + dm;
    } else {
       m1[i] = -col11[star[i]] + mv[star[i]] + dm;
    }
  }
  for (i = 0; i < nstar; i++) {
    if (band2 == 'U') {
      m2[i] = col21[star[i]] + col22[star[i]] + mv[star[i]] + dm;
    } else if (band2 == 'B') {
      m2[i] = col21[star[i]] + mv[star[i]] + dm;
    } else if (band2 == 'V') {
      m2[i] = mv[star[i]] + dm;
    } else if (band2 == 'H') {
      m2[i] = col21[star[i]] - col22[star[i]] + mv[star[i]] + dm;
    } else {
      m2[i] = -col21[star[i]] + mv[star[i]] + dm;
    }
  }

  /* Generate positions for the stars */
  c1 = 0.5 * (ceil(fov*1000.0 / ps1) + 1.0);
  c2 = 0.5 * (ceil(fov*1000.0 / ps2) + 1.0);
  for (i = 0; i < nstar; i++) {
    x1[i] = 0.5 + ceil(fov*1000.0 / ps1) * gsl_rng_uniform(rng);
    y1[i] = 0.5 + ceil(fov*1000.0 / ps1) * gsl_rng_uniform(rng);
    x2[i] = (x1[i] - c1) * ps1/ps2 + c2;
    y2[i] = (y1[i] - c1) * ps1/ps2 + c2;    
  }

  /* Output */

  /* Header */
  printf("# spf = %s  DM = %.1f  SB = %.1f mag/arsec^2  \
FoV size = %.3f arcsec  band1 = %c  band2 = %c  pixel size 1 = %.4f mas  \
pixel size 2 = %.4f  seed = %ld\n#\n",
	 spf, dm, mu, fov, band1, band2, ps1, ps2, seed);

  i = j = 0;
  while (fgets(buffer, L_WORD, data_file) != NULL) {
    if (buffer[0] == '#') {
      if ((buf = wdinstr("weight", buffer, &brcol)) != NULL) {
	*buf = '\0';
	printf("%s %c      %c      %c-%c        x_%c      y_%c      x_%c      y_%c\n",
	       buffer, band1, band2, band2, band1, band1, band1, band2, band2);
	brcol--;
      } else
	printf("%s", buffer);
    } else {
      buf = mv2col(buffer, brcol);
      *buf = '\0';
      while (star[i] == j) {
	printf("%s%6.3f %6.3f %6.3f  %8.2f %8.2f %8.2f %8.2f\n", buffer, m1[i],
	       m2[i], m2[i] - m1[i], x1[i], y1[i], x2[i], y2[i]);
	i++;
      }
      j++;
    }
  }

  /* Clean up */
  fclose(data_file);
  gsl_rng_free(rng);
  free(mv);
  if (col11 != NULL)
    free(col11);
  if (col12 != NULL)
    free(col12);
  if (col21 != NULL)
    free(col21);
  if (col22 != NULL)
    free(col22);
  free(w);
  free(star);
  free(m1);
  free(m2);
  free(x1);
  free(y1);
  free(x2);
  free(y2);

  return 0;
}
