/******************************************************************************
* Make a catalogue of objects for spectroscopy case from a stellar population
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#define  L_WORD   1000  /* Longest string that can be handled */


char      *progname;


/******************************************************************************
* Print the usage message
******************************************************************************/
 
void usage(void) {

  fprintf(stderr, "\nUsage: %s [options]\n", progname);
  fprintf(stderr, "\nMake a catalogue of objects for the spectroscopy case\n\
from a given stellar population (produced by\n\
rspmkstelpop_basti). Any parameters not specified on\n\
the command line will be queried for.\n");
  fprintf(stderr, "\nOptions include:\n \
\t-help\t\tprint this message\n \
\t-spf file\tstellar population file\n\
\t-dm DM\t\tdistance modulus\n\
\t-sb mu\t\tV band surface brightness\n\
\t-spax x\t\tspaxel size in mas\n\
\t-vsig s\t\tvelocity dispersion of the stars\n\
\t-seed r\t\trandom number seed\n\
\n");
  exit(3);
}


/******************************************************************************
* Main program
******************************************************************************/

int main(int argc, char *argv[]) {

  int        i, j, npts, *star, nstar, brcol;
  unsigned long int seed;
  float      dm, mu, ps, velsig, *mv, *vel;
  double     reqflux, tflux, *w, maxw;
  char       spf[L_WORD], buffer[L_WORD], *buf;
  FILE       *data_file;
  gsl_rng    *rng;
  float      *readafcol(char *, char *, int *);
  double     *readadcol(char *, char *, int *);
  char       *wdinstr(char *, char *m, int *), *mv2col(char *, int);
  FILE       *faskropen(char *, char *, int);
  void       errormsg(char *, ...), get_input(char *, char *, ...);


  progname = strrchr(argv[0], '/');
  progname = (progname == NULL) ? argv[0] : progname + 1;
  
  /* Look for -help switch */
  for (i = 1; i < argc; i++)
    if (!strcmp(argv[i], "-help"))
      usage();

  /* Some setup */
  strcpy(spf, " ");
  dm = -1.0;
  mu = -1.0;
  ps = -1.0;
  velsig = -1.0;
  seed = 0;

  /* Look for other switches */
  for (i = 1; i < argc-1; i++)
    if (!strcmp(argv[i], "-spf")) {
      strcpy(spf, argv[++i]);
    } else if (!strcmp(argv[i], "-dm")) {
      if (sscanf(argv[++i], "%f", &dm) != 1 || dm < 0.0)
	errormsg("Bad distance modulus code: %s", argv[i]);
    } else if (!strcmp(argv[i], "-sb")) {
      if (sscanf(argv[++i], "%f", &mu) != 1 || mu < 0.0)
	errormsg("Bad surface brightness: %s", argv[i]);
    } else if (!strcmp(argv[i], "-spax")) {
      if (sscanf(argv[++i], "%f", &ps) != 1 || ps < 0.0)
	errormsg("Bad spaxel size: %s", argv[i]);
    } else if (!strcmp(argv[i], "-vsig")) {
      if (sscanf(argv[++i], "%f", &velsig) != 1 || velsig < 0.0)
	errormsg("Bad velocity dispersion: %s", argv[i]);
    } else if (!strcmp(argv[i], "-seed")) {
      if (sscanf(argv[++i], "%lu", &seed) != 1)
	errormsg("Bad random number seed: %s", argv[i]);
    } else
      errormsg("Unknown command line option: %s", argv[i]);


  /* Get all missing input */

  /* Stellar population file */
  if (!strcmp(spf, " "))
    i = 1;
  else
    i = 3;
  if ((data_file = faskropen("\nStellar population file?", spf, i)) == NULL)
    errormsg("Couldn't open %s!", spf);

  /* Distance modulus */
  if (dm < 0.0) {
    dm = 31.2;
    do
      get_input("\nDistance modulus?", "%.1f", &dm);
    while (dm < 0.0);
  }

  /* Surface brightness */
  if (mu < 0.0) {
    mu = 22.0;
    do
      get_input("\nSurface brightness (in mag/arcsec^2)?", "%.1f", &mu);
    while (mu < 0.0);
  }

  /* Spaxel size */
  if (ps < 0.0) {
    ps = 50.0;
    do
      get_input("\nSpaxel size (in mas)?", "%.1f", &ps);
    while (ps < 0.0);
  }

  /* Velocity dispersion */
  if (velsig < 0.0) {
    velsig = 100.0;
    do
      get_input("\nVelocity dispersion of stars (in km/s)?", "%.1f", &velsig);
    while (velsig < 0.0);
  }

  /* Random number seed */
  if (seed == 0)
    seed = time((time_t *)0);

  /* Initialise the random number generator */
  rng = gsl_rng_alloc(gsl_rng_taus2);
  gsl_rng_set(rng, seed);
  fprintf(stderr, "\nRandom number seed = %lu\n", seed);

  /* Read in mags and weights */
  mv = readafcol("Mv", spf, &npts);
  w = readadcol("weight", spf, &npts);
  
  /* Requested flux */
  reqflux = pow(10.0, -0.4 * mu) * ps*ps * 1.0e-6;

  /* Total flux from the stellar population */
  tflux = 0.0;
  for (i = 0; i < npts; i++)
    tflux += pow(10.0, -0.4 * (mv[i] + dm)) * w[i];

  /* Required number of stars */
  nstar = (int)rint(reqflux / tflux);
  fprintf(stderr, "Number of stars = %d\n", nstar);

  /* Make some memory */
  if ((star = (int *)calloc(nstar, sizeof(int))) == NULL)
    errormsg("Couldn't allocate memory!");
  if ((vel = (float *)calloc(nstar, sizeof(float))) == NULL)
    errormsg("Couldn't allocate memory!");

  /* Monte Carlo the stars */
  maxw = 1.001 * gsl_stats_max(w, 1, npts);
  for (i = 0; i < nstar; i++) {
    do
      star[i] = gsl_rng_uniform_int(rng, npts);
    while (maxw * gsl_rng_uniform(rng) > w[star[i]]);
  }

  /* Sort the stars; don't add any star properties before this */
  gsl_sort_int(star, 1, nstar);

  /* Actual surface brightness produced by the stars */
  tflux = 0.0;
  for (i = 0; i < nstar; i++)
    tflux += pow(10.0, -0.4 * (mv[star[i]] + dm));
  tflux = -2.5 * log10(tflux / (ps*ps*1.0e-6));
  fprintf(stderr, "Actual surface brightness = %.3f mag/arcsec^2\n", tflux);

  /* Generate velocities for the stars */
  for (i = 0; i < nstar; i++)
    vel[i] = gsl_ran_gaussian(rng, velsig);

  /* Output */

  /* Header */
  printf("# spf = %s   DM = %.1f   SB = %.1f mag/arsec^2   \
spax size = %.1f mas   vsig = %.1f km/s   seed = %ld\n#\n", spf, dm, mu, ps,
	 velsig, seed);

  i = j = 0;
  while (fgets(buffer, L_WORD, data_file) != NULL) {
    if (buffer[0] == '#') {
      if ((buf = wdinstr("weight", buffer, &brcol)) != NULL) {
	*buf = '\0';
	printf("%s    v\n", buffer);
	brcol--;
      } else
	printf("%s", buffer);
    } else {
      buf = mv2col(buffer, brcol);
      *buf = '\0';
      while (star[i] == j) {
	printf("%s%7.1f\n", buffer, vel[i]);
	i++;
      }
      j++;
    }
  }

  /* Clean up */
  fclose(data_file);
  gsl_rng_free(rng);
  free(mv);
  free(w);
  free(star);
  free(vel);

  return 0;
}
