/******************************************************************************
* Make a synthetic stellar population
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "const.h"

#define  L_WORD   1000  /* Longest string that can be handled */
#define  ISO_PATH "~/stelpop_code/isochrones/basti/"
                        /* Path to the isochrones*/
#define  N_T      55    /* Number of available timesteps */


char      *progname;


/******************************************************************************
* Print the usage message
******************************************************************************/
 
void usage(void) {

  fprintf(stderr, "\nUsage: %s [options]\n", progname);
  fprintf(stderr, "\nMake a synthetic stellar population. Run %s\n\
interactively (i.e. without any options) to find out the allowed\n\
values of the various codes below. If the IMF slope is given on\n\
the command line but the IMF mass range is not, then a default\n\
value is assumed for the latter.\n\n\
When the SFH is specified in a file the format is:\n\
t in Gyr       relative SFR\n", progname);
  fprintf(stderr, "\nOptions include:\n \
\t-help\t\t\tprint this message\n \
\t-ae code\t\tscaled solar or alpha-enhanced models\n\
\t-comp code\t\tchemical composition\n\
\t-imf alpha\t\tIMF slope\n\
\t-mr M_min M_max\t\tminimum and maximum masses of IMF\n\
\t-sfh code params\tSFH\n\
\n");
  exit(3);
}


/******************************************************************************
* Main program
******************************************************************************/

int main(int argc, char *argv[]) {

  int       i, j, k, ae, ccomp, sfh, flag, npts, tnpts, brcol;
  float     alpha, minmass, maxmass, tstart, tend, tau, tdum, sfrdum;
  float     t[N_T]={30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0, 150.0, \
		    200.0, 250.0, 300.0, 350.0, 400.0, 450.0, 500.0, 600.0, \
		    700.0, 800.0, 900.0, 1000.0, 1250.0, 1500.0, 1750.0, \
		    2000.0,2250.0, 2500.0, 2750.0, 3000.0, 3250.0, 3500.0, \
		    3750.0, 4000.0, 4500.0, 5000.0, 5500.0, 6000.0, 6500.0, \
		    7000.0, 7500.0, 8000.0, 8500.0, 9000.0, 9500.0, 10000.0, \
		    10500.0, 11000.0, 11500.0, 12000.0, 12500.0, 13000.0, \
		    13500.0, 14000.0, 14500.0, 15000.0};
  double    sfr[N_T], *imf, *w, *logg, *massin, *mass, *logl, *logt, sum, tsum;
  char      *buf, buffer[L_WORD], buffer2[L_WORD], isofn[L_WORD],
            isofnf[L_WORD], sfrfn[L_WORD];
  FILE      *data_file;
  int       idxnval(float *, int, float), nlines(FILE *);
  char      *wdinstr(char *, char *, int *), *mv2col(char *, int);
  FILE      *faskropen(char *, char *, int);
  void      errormsg(char *, ...), twid2home(char *, char *),
            get_input(char *, char *, ...);


  progname = strrchr(argv[0], '/');
  progname = (progname == NULL) ? argv[0] : progname + 1;
  
  /* Look for -help switch */
  for (i = 1; i < argc; i++)
    if (!strcmp(argv[i], "-help"))
      usage();

  /* Some setup */
  ae = 0;
  ccomp = 0;
  alpha = 999.0;
  minmass = 0.5;
  maxmass = 11.0;
  sfh = 0;

  /* Look for other switches */
  for (i = 1; i < argc-1; i++)
    if (!strcmp(argv[i], "-ae")) {
      if (sscanf(argv[++i], "%d", &ae) != 1 || ae < 1 || ae > 2)
	errormsg("Bad alpha enhancement code: %s", argv[i]);
    } else if (!strcmp(argv[i], "-comp")) {
      if (sscanf(argv[++i], "%d", &ccomp) != 1 || ccomp < 1 || ccomp > 11)
	errormsg("Bad chemical composition code: %s", argv[i]);
    } else if (!strcmp(argv[i], "-imf")) {
      if (sscanf(argv[++i], "%f", &alpha) != 1)
	errormsg("Bad IMF slope: %s", argv[i]);
    } else if (!strcmp(argv[i], "-mr")) {
      if (sscanf(argv[++i], "%f", &minmass) != 1 || minmass < 0.5 
	  || minmass > 11.0)
	errormsg("Bad minimum mass: %s", argv[i]);
      if (sscanf(argv[++i], "%f", &maxmass) != 1 || maxmass < 0.5 
	  || maxmass > 11.0 || maxmass <= minmass)
	errormsg("Bad minimum mass: %s", argv[i]);
    } else if (!strcmp(argv[i], "-sfh")) {
      if (sscanf(argv[++i], "%d", &sfh) != 1 || sfh < 1 || sfh > 3)
	errormsg("Bad SFH code: %s", argv[i]);
      if (sfh == 1) {
	if (sscanf(argv[++i], "%f", &tstart) != 1 || tstart < 0.0 || 
	    tstart > 15.0)
	  errormsg("Bad t_start: %s", argv[i]);
	if (sscanf(argv[++i], "%f", &tend) != 1 || tend < 0.0 || 
	    tend > 15.0 || tstart < tend)
	  errormsg("Bad t_end: %s", argv[i]);
      } else if (sfh == 2) {
	if (sscanf(argv[++i], "%f", &tstart) != 1 || tstart < 0.0 || 
	    tstart > 15.0)
	  errormsg("Bad t_start: %s", argv[i]);
	if (sscanf(argv[++i], "%f", &tau) != 1 || tau < 0.0)
	  errormsg("Bad tau: %s", argv[i]);
      } else if (sfh == 3) {
	strcpy(sfrfn, argv[++i]);
      } else
	errormsg("Something is very wrong!");

    } else
      errormsg("Unknown command line option: %s", argv[i]);

  /* More setup */
  sum = 0.0;
  tsum = 0.0;
  twid2home(isofn, ISO_PATH);
  for (i = 0; i < N_T; i++)
    sfr[i] = 0.0;


  /* Alpha enhancement */

  if (ae == 0) {

    /* Default */
    ae = 1;

    do
      get_input("\nSelect scaled solar (1) or alpha enhanced (2) models:",
		"%d", &ae);
    while (ae < 1 || ae > 2);
  }

  if (ae == 1)
    strcat(isofn, "scaled_solar/");
  else if (ae == 2)
    strcat(isofn, "alpha_enhanced/");
  else
    errormsg("Something is very wrong!");


  /* Chemical composition */

  if (ccomp == 0) {
    /* Get the chemical composition */

    /* Default */
    ccomp = 9;

    /* Open the file containing the options */
    sprintf(buffer, "%smenu.dat", isofn);
    if ((data_file = faskropen("Menu file?", buffer, 3)) == NULL)
      errormsg("Couldn't open %s!", buffer);

    /* Display the options */
    fprintf(stderr, "\n");
    while (fgets(buffer, L_WORD, data_file) != NULL) {
      buf = &buffer[strlen(buffer)-2];
      while (!isspace(*buf) && buf != buffer)
	buf--;
      *buf = '\n';
      buf++;
      *buf = '\0';
      fprintf(stderr, "%s", buffer);
    }
    fclose(data_file);

    /* Get the user's choice */
    do
      get_input("\nSelect chemical composition from above list:", "%d",
		&ccomp);
    while (ccomp < 1 || ccomp > 11);
  }

  /* Get the root file appropriate for the chosen chemical composition */
  /* Open the file containing the options */
  sprintf(buffer, "%smenu.dat", isofn);
  if ((data_file = faskropen("Menu file?", buffer, 3)) == NULL)
    errormsg("Couldn't open %s!", buffer);

  flag = 0;
  while (fgets(buffer, L_WORD, data_file) != NULL)
    if (buffer[0] != '#') {
      if (sscanf(buffer, "%d %*s %*s %*s %s", &i, buffer2) != 2)
	errormsg("Having trouble reading the menu file!");
      if (i == ccomp) {
	strcat(isofn, buffer2);
	flag = 1;
	break;
      }
    }
  fclose(data_file);
  if (flag == 0)
    errormsg("Something is very wrong!");


  /* IMF */

  if (alpha > 998.0) {
    /* Get the IMF slope */

    /* Default */
    alpha = -2.35;

    get_input("\nSlope of the IMF?", "%.2f", &alpha);

    do
      get_input("\nMinimum and maximum mass of IMF (in M_sol)?", "%.3f %.3f",
		&minmass, &maxmass);
    while (minmass < 0.5 || minmass > 11.0 || maxmass < 0.5 
	   || maxmass > 11.0 || maxmass <= minmass);
  }


  /* SFH */

  if (sfh == 0) {
    /* Get SFH */

    /* Defaults */
    sfh = 1;
    tstart = 14.0;
    tend = 5.0;
    tau = 1.0;

    /* Menu */
    fprintf(stderr, "\n# Option   SFH = SFR(t)\n");
    fprintf(stderr, "  1        constant\n");
    fprintf(stderr, "  2        exponentially decaying\n");
    fprintf(stderr, "  3        from file\n");

    do
      get_input("\nSelect SFH from above list:", "%d", &sfh);
    while (sfh < 1 || sfh > 3);

    /* Get SFH parameters */
    if (sfh == 1) {
      do 
	get_input("Start and end time (in Gyr)?", "%.3f %.3f", &tstart, &tend);
      while (tstart < 0.0 || tstart > 15.0 || tend < 0.0 || tend > 15.0 
	     || tstart <= tend);
      
    } else if (sfh == 2) {
      do 
	get_input("Start and decay time (in Gyr)?", "%.3f %.3f", &tstart,
		  &tau);
      while (tstart < 0.0 || tstart > 15.0 || tau < 0.0);

    } else if (sfh == 3) {
      get_input("SFR file name?", "!%s", sfrfn);

    } else
      errormsg("Something is very wrong!");
  }

  /* Calculate the inter-isochrone, i.e. SFR weights */
  if (sfh == 1) {
    tstart *= 1000.0;
    tend *= 1000.0;

    sum = 0.0;
    for (i = 0; i < N_T; i++) {
      if (t[i] >= tend && t[i] <= tstart)
	sfr[i] = 1.0;
      sum += sfr[i];
    }

    if (sum == 0.0) {
      i = idxnval(t, N_T, tend);
      if (i == 0)
	tend = t[i];
      else
	tend = 0.5 * (t[i-1] + t[i]);
      sum = 0.0;
      for (i = 0; i < N_T; i++) {
	if (t[i] >= tend && t[i] <= tstart)
	  sfr[i] = 1.0;
	sum += sfr[i];
      }
    }

  } else if (sfh == 2) {
    tstart *= 1000.0;
    tau *= 1000.0;

    sum = 0.0;
    for (i = 0; i < N_T; i++) {
      if (t[i] <= tstart)
	sfr[i] = exp(-(tstart - t[i]) / tau);
      sum += sfr[i];
    }

  } else if (sfh == 3) {
    /* Open the SFR file */
    if ((data_file = faskropen("SFR file?", sfrfn, 3)) == NULL)
      errormsg("Couldn't open %s!", sfrfn);
    
    sum = 0.0;
    while (fgets(buffer, L_WORD, data_file) != NULL)
      if (buffer[0] != '#') {
	if (sscanf(buffer, "%f %f", &tdum, &sfrdum) != 2)
	  errormsg("Having trouble reading the data from %s!", sfrfn);
	tdum *= 1000.0;
	i = idxnval(t, N_T, tdum);
	sfr[i] = sfrdum;
	sum += sfr[i];
      }
    fclose(data_file);
    
  } else
    errormsg("Something is very wrong!");

  /* Normalise the SFR weights */
  for (i = 0; i < N_T; i++)
    sfr[i] /= sum;


  /* Loop over isochrone files and calculate logg and weights */
  k = 0;
  tnpts = 0;
  logg = NULL;
  w = NULL;
  tsum = 0.0;
  for (i = 0; i < N_T; i++) {
    if (sfr[i] > 0.0) {

      /* Open the isochrone file */
      sprintf(isofnf, "%s.t6%05.0f", isofn, t[i]);
      if ((data_file = faskropen("Isochrone file?", isofnf, 3)) == NULL)
	errormsg("Couldn't open %s!", isofnf);

      npts = nlines(data_file);

      /* Make some memory */
      if ((imf = (double *)calloc(npts, sizeof(double))) == NULL)
	errormsg("Couldn't allocate memory!");
      if ((massin = (double *)calloc(npts, sizeof(double))) == NULL)
	errormsg("Couldn't allocate memory!");
      if ((mass = (double *)calloc(npts, sizeof(double))) == NULL)
	errormsg("Couldn't allocate memory!");
      if ((logl = (double *)calloc(npts, sizeof(double))) == NULL)
	errormsg("Couldn't allocate memory!");
      if ((logt = (double *)calloc(npts, sizeof(double))) == NULL)
	errormsg("Couldn't allocate memory!");

      /* Calculate logg and the intra-isochrone, i.e. IMF weights */
      j = 0;
      sum = 0.0;
      while (j < npts && fgets(buffer, L_WORD, data_file) != NULL)
	if (buffer[0] != '#') {
	  if (sscanf(buffer, "%lf %lf %lf %lf", &massin[j], &mass[j], &logl[j],
		     &logt[j]) != 4)
	    errormsg("Having trouble reading the data from %s!", isofnf);

	  /* If this is the first line, go straight to the next */
	  if (j > 0) {
	    if (massin[j] <= massin[j-1])
	      errormsg("The isochrone is not sorted by mass!");

	    if (j == 1) {
	      /* Do the first one */
	      if (massin[j-1] >= minmass && massin[j-1] <= maxmass) {
		imf[j-1] = pow(massin[j-1], alpha) * (massin[j] - massin[j-1]);
	      } else
		imf[j-1] = 0.0;
	      sum += imf[j-1];
	    }

	    if (massin[j] >= minmass && massin[j] <= maxmass) {
	      imf[j] = pow(massin[j], alpha) * (massin[j] - massin[j-1]);
	    } else
	      imf[j] = 0.0;
	    sum += imf[j];
	  }
	  j++;
	}
      npts = j;
      tnpts += npts;
      fclose(data_file);
      
      /* Normalise the IMF weights */
      if (sum > 0.0)
	for (j = 0; j < npts; j++)
	  imf[j] /= sum;

      /* Calculate logg and total weights */
      if ((logg = (double *)realloc(logg, tnpts * sizeof(double))) == NULL)
	errormsg("Couldn't reallocate memory!");
      if ((w = (double *)realloc(w, tnpts * sizeof(double))) == NULL)
	errormsg("Couldn't reallocate memory!");

      for (j = 0; j < npts; j++) {
	logg[k] = log10(G * 4.0*PI * S_B) + log10(mass[j] * M_S) 
	  - logl[j] - log10(L_S) + 4.0*logt[j] + 2.0;

	w[k] = sfr[i] * imf[j];
	tsum += w[k];
	k++;
      }

      /* Clean up */
      free(imf);
      free(massin);
      free(mass);
      free(logl);
      free(logt);
    }
  }
  
  if (k != tnpts)
    errormsg("Something is very wrong!");

  /* Normalize the final weights */
  if (tsum > 0.0)
    for (k = 0; k < tnpts; k++)
      w[k] /= tsum;


  /* Loop over isochrone files again and write output */

  /* Header */
  fprintf(stderr, "\n");
  printf("# imf slope = %.2f   ", alpha);
  if (sfh == 1)
    printf("SFR = constant, t_start = %.3f, t_end = %.3f\n", tstart/1000.0,
	   tend/1000.0);
  else if (sfh == 2)
    printf("SFR = exponential, t_start = %.3f, tau = %.3f\n", tstart/1000.0,
	   tau/1000.0);
  else if (sfh == 3)
    printf("SFR = from file %s\n", sfrfn);
  else
    errormsg("Something is very wrong!");
  printf("#\n# Header from first isochrone:\n");

  flag = 1;
  k = 0;
  for (i = 0; i < N_T; i++) {
    if (sfr[i] > 0.0) {

      /* Open the isochrone file */
      sprintf(isofnf, "%s.t6%05.0f", isofn, t[i]);
      if ((data_file = faskropen("Isochrone file?", isofnf, 3)) == NULL)
	errormsg("Couldn't open %s!", isofnf);

      /* Output */
      while (fgets(buffer, L_WORD, data_file) != NULL)
	if (buffer[0] == '#') {
	  /* Header */
	  if (flag) {
	    if ((buf = wdinstr("Mv", buffer, &brcol)) != NULL) {
	      /* Insert logg */
	      buffer[strlen(buffer)-1] = '\0';
	      buf -= 4;
	      *buf = '\0';
	      printf("%slogg        %s      age     weight\n", buffer, buf+4);
	      brcol--;
	    } else
	      printf("%s", buffer);
	  }

	} else {
	  if (w[k] > 0.0) {
	    buffer[strlen(buffer)-1] = '\0';
	    buf = mv2col(buffer, brcol);
	    if (*buf == '.')
	      buf -= 3;
	    else if (*buf == '-' && *(buf+1) != '.')
	      buf -= 1;
	    else 
	      buf -= 2; 
	    *buf = '\0';
	    printf("%s%.5f  %s   %7.1f   %e\n", buffer, logg[k], buf+1, t[i],
		   w[k]);
	  }
	  k++;
	}
      fclose(data_file);
      flag = 0;
    }
  }

  /* Clean up */
  free(logg);
  free(w);

  return 0;
}
